/*

	kipdj.c

	Kaplan Internet Protocols:
	Interface DJGPP / Windows Sockets

	(C) Angel Ortega 1997/1998


	Basado en mltiples fuentes y docs trados de Internet.

	Faltan:

	- DNS.

*/


#include <stdio.h>
#include <dpmi.h>
#include <pc.h>
#include <string.h>
#include <sys/farptr.h>
#include <sys/movedata.h>
#include <sys/segments.h>
#include <errno.h>

#include "kipdj.h"


/* constantes para select */

#define FD_READ 	0x01
#define FD_WRITE	0x02
#define FD_OOB		0x04
#define FD_ACCEPT	0x08
#define FD_CONNECT	0x10
#define FD_CLOSE	0x20


/*******************
	Datos
********************/

/* activos s/no */
int _kipdj_up=0;

/* ltimo error de winsock */
int _kipdj_error=0;

/* flag para activacin de funciones bloqueantes */
int _kipdj_blocking=1;


/* direccin del VxD cargador de VxD,s. */
static int _vxd_entry_point[2];

/* direccin del winsock VxD */
static int _winsock_entry_point[2];

/* descriptor del bloque de memoria para parmetros */
static int _winsock_param;

/* descriptor del bloque de memoria para datos */
static int _winsock_data;

/* defines para simplificar accesos */

#define _movedata_to(sel,dat,len)   movedata(_my_ds(),(int)dat,sel,0,len)
#define _movedata_from(sel,dat,len) movedata(sel,0,_my_ds(),(int)dat,len)


/********************
	Cdigo
*********************/


static void _vxd_get_entry(int * entry_point, int device_id)
/* obtiene la direccin de entrada de un VxD */
{
	entry_point[0]=0;
	entry_point[1]=0;

	asm ("pushl   %%es	      \n\
	      movw    %%di, %%es      \n\
	      intb    $0x2f	      \n\
	      movl    $0, %%ecx       \n\
	      movw    %%es, %%cx      \n\
	      popl    %%es"
		: "=c" (entry_point[1]), "=D" (entry_point[0])
		: "a" (0x1684), "b" (device_id), "D" (0)
		: "%eax", "%edx");
}


static void _winsock_entry(int func)
/* llama a la funcin func de wsock.vxd */
{
	_kipdj_error=-1;

	asm volatile ("pushl   %%es		       \n\
		       pushl   %%ecx		       \n\
		       popl    %%es		       \n\
		       lcall   __winsock_entry_point   \n\
		       andl    %%eax, 0x0000ffff       \n\
		       popl    %%es"
		 : "=a" (_kipdj_error)
		 : "a" (func), "b" (0), "c" (_winsock_param));

	/* si _kipdj_error es 0xffff, try again */
}


static int _vxd_load_device(char * devname)
/* carga un VxD */
{
	int err;
	int dev_desc;
	__dpmi_meminfo dpmi_dev;

	dpmi_dev.handle=0;
	dpmi_dev.size=strlen(devname)+1;
	dpmi_dev.address=0;

	__dpmi_allocate_memory(&dpmi_dev);
	dev_desc=__dpmi_allocate_ldt_descriptors(1);
	__dpmi_set_segment_base_address(dev_desc,dpmi_dev.address);
	__dpmi_set_segment_limit(dev_desc,dpmi_dev.size);

	_movedata_to(dev_desc, devname, dpmi_dev.size);

	/* el valor 1 en "a" es LoadDevice */
	asm ("pushl   %%ds	      \n\
	      movw    %%cx, %%ds      \n\
	      lcall   %%cs:__vxd_entry_point \n\
	      setc    %%al	      \n\
	      movzbl  %%al, %%eax     \n\
	      popl    %%ds"
		: "=a" (err)
		: "a" (1), "c" (dev_desc), "d" (0));

	__dpmi_free_ldt_descriptor(dev_desc);
	__dpmi_free_memory(dpmi_dev.handle);

	/* si se ha podido cargar, se activan */
	if(err!=0)
		_kipdj_up=1;

	return(err);
}


static void _call_winsock(int func, void * params, int paramsize)
/* llama a winsock, rellenando los parmetros */
{
	/* si no funciona, se devuelve un error
	   de WSA_NOTINITIALIZED */
	if(!_kipdj_up)
	{
		_kipdj_error=10093;
		return;
	}

	_movedata_to(_winsock_param,params,paramsize);

	_winsock_entry(func);

	_movedata_from(_winsock_param,params,paramsize);
}


/** funciones de sockets **/

static int _kipdj_wait_for_event(int s, int event)
/* variante de select, para esperar eventos en sockets */
{
	int to;
	struct
	{
		void * readlist;
		void * writelist;
		void * exceptlist;
		int readcount;
		int writecount;
		int exceptcount;
		void * routine;
		int context;
		struct
		{
			long iostatus;
			char iocompleted;
			char iocancelled;
			char iotimedout;
			char iospare1;
		} ios;
	} params;
	struct
	{
		void * socket;
		long events;
		long context;
	} sock_list;

	memset(&params,0,sizeof(params));
	memset(&sock_list,0,sizeof(sock_list));

	params.readlist=(void *)((_winsock_data<<16));
	params.writelist=(void *)((_winsock_data<<16));
	params.exceptlist=(void *)((_winsock_data<<16));

	sock_list.socket=(void *)s;
	sock_list.events=event;
	sock_list.context=0;

	params.readcount=1;
	params.writecount=1;
	params.exceptcount=1;
	params.routine=(void *)(-1);

	/* context apunta a ios */
	params.context=(_winsock_param<<16)+
		(sizeof(void *)*4)+(sizeof(int)*4);

	_movedata_to(_winsock_data, &sock_list, sizeof(sock_list));

	_call_winsock(0x10a, &params, sizeof(params));

	if(_kipdj_error && _kipdj_error!=0xffff)
		return(-1);

	/* espera a que el flag iocompleted est activado */
	for(to=0x3000000;!params.ios.iocompleted && to;to--)
		_movedata_from(_winsock_param, &params, sizeof(params));

	if(to==0)
	{
		/* timeout (puede ser tambin ocasionado por
		   un conection refused, pero no s cmo detectarlo) */
		_kipdj_error=10060;
		return(-1);
	}

	/* ya est: limpiar el select */

	params.readlist=(void *)((_winsock_data<<16));
	params.writelist=(void *)((_winsock_data<<16));
	params.exceptlist=(void *)((_winsock_data<<16));

	sock_list.socket=(void *)s;
	sock_list.events=event;
	sock_list.context=0;

	params.readcount=1;
	params.writecount=1;
	params.exceptcount=1;

	_call_winsock(0x10b, &params, sizeof(params));

	if(to==0)
		return(-1);

	return(0);
}


int socket(int domain, int type, int protocol)
/* crea un socket */
{
	struct
	{
		int domain;
		int type;
		int protocol;
		void * new_socket;
		int socket;
	} params;

	params.domain=domain;
	params.type=type;
	params.protocol=protocol;
	params.new_socket=0;
	params.socket=0;

	_call_winsock(0x110,&params,sizeof(params));

	if(_kipdj_error)
		return(-1);

	return((int) params.new_socket);
}


int closesocket(int s)
/* cierra el socket */
{
	struct
	{
		void * socket;
	} params;

	params.socket=(void *) s;

	_call_winsock(0x102,&params,sizeof(params));

	if(_kipdj_error)
		return(-1);

	return(0);
}


int connect(int s, struct sockaddr * addr, int len)
/* conecta con un server */
{
	struct
	{
		void * address;
		void * socket;
		long len;
		void * routine;
		long context;
		struct sockaddr addr;
	} params;

	/* rellena la estructura */
	params.socket=(void *) s;
	params.len=len;
	params.routine=0;
	params.context=0;

	/* hace que address apunte justo sobre params.addr */
	params.address=(void *)((_winsock_param << 16)+
		(sizeof(void *)*3)+(sizeof(long)*2));

	/* copia la estructura addr */
	memcpy(&params.addr, addr, len);

	_call_winsock(0x103, &params, sizeof(params));

	/* recupera la estructura addr */
	memcpy(addr, &params.addr, len);

	if(_kipdj_error && _kipdj_error!=0xffff)
		return(-1);

	if(_kipdj_blocking)
		return(_kipdj_wait_for_event(s, FD_CONNECT));

	return(0);
}


int sendto(int s, void * data, int len, unsigned int flags,
	struct sockaddr * to, int tolen)
/* enva a la direccin especificada */
{
	struct
	{
		void * buffer;
		void * address;
		void * socket;
		int len;
		int flags;
		int addrlen;
		int sent;
		void * routine;
		void * context;
		int timeout;
		struct sockaddr addr;
	} params;

	params.socket=(void *) s;
	params.len=len;
	params.flags=flags;
	params.sent=0;
	params.routine=NULL;
	params.context=0;
	params.timeout=0;

	if(to==NULL)
	{
		params.addrlen=0;
		params.address=NULL;
	}
	else
	{
		params.addrlen=tolen;
		params.address=(void *)((_winsock_param << 16)+
			(sizeof(void *)*5)+(sizeof(int)*5));

		/* se copia la informacin del host destino */
		memcpy(&params.addr,to,tolen);
	}

	/* direccin del buffer */
	params.buffer=(void *) (_winsock_data << 16);

	/* rellena el buffer */
	_movedata_to(_winsock_data,data,len);

	_call_winsock(0x10d,&params,sizeof(params));

	if(_kipdj_error)
		return(-1);

	return(params.sent);
}


int send(int s, void * data, int len, unsigned int flags)
/* enva datos por el socket */
{
	return(sendto(s,data,len,flags,NULL,0));
}


int recvfrom(int s, void * data, int len, unsigned int flags,
	struct sockaddr * from, int * fromlen)
/* recibe y rellena informacin sobre el origen */
{
	struct
	{
		void * buffer;
		void * address;
		void * socket;
		int len;
		int flags;
		int addrlen;
		int received;
		void * routine;
		void * context;
		int timeout;
		struct sockaddr addr;
	} params;

	if(_kipdj_blocking)
		_kipdj_wait_for_event(s, FD_READ);

	params.socket=(void *) s;
	params.len=len;
	params.flags=flags;
	params.received=0;
	params.routine=NULL;
	params.context=0;
	params.timeout=0;

	if(from==NULL)
	{
		params.addrlen=0;
		params.address=NULL;
	}
	else
	{
		params.addrlen=*fromlen;
		params.address=(void *)((_winsock_param << 16)+
			(sizeof(void *)*5)+(sizeof(int)*5));
	}

	/* direccin del buffer */
	params.buffer=(void *) (_winsock_data << 16);

	_call_winsock(0x109, &params, sizeof(params));

	/* recupera el buffer */
	_movedata_from(_winsock_data,data,params.received);

	if(_kipdj_error)
	{
		if(_kipdj_error==0xffff)
			errno=EWOULDBLOCK;
		else
			errno=EIO;

		return(-1);
	}

	if(from!=NULL)
	{
		*fromlen=params.addrlen;
		memcpy(from, &params.addr, *fromlen);
	}

	return(params.received);
}


int recv(int s, void * data, int len, unsigned int flags)
/* recibe desde el socket */
{
	return(recvfrom(s,data,len,flags,NULL,NULL));
}


int listen(int s, int backlog)
/* activa la aceptacin de conexiones en el socket */
{
	struct
	{
		void * socket;
		int backlog;
	} params;

	params.socket=(void *)s;
	params.backlog=backlog;

	_call_winsock(0x108, &params, sizeof(params));

	if(_kipdj_error && _kipdj_error!=0xffff)
		return(-1);

	return(0);
}


int accept(int s, struct sockaddr * addr, int * addrlen)
/* acepta conexiones */
{
	struct
	{
		void * address;
		void * list_socket;
		void * conn_socket;
		int len;
		int socket_handle;
		void * routine;
		int context;
		struct sockaddr addr;
	} params;

	/* espera a que haya una conexin */
	_kipdj_wait_for_event(s, FD_ACCEPT);

	params.list_socket=(void *)s;
	params.conn_socket=NULL;
	params.len=*addrlen;
	params.socket_handle=0;
	params.routine=NULL;
	params.context=0;

	params.address=(void *) ((_winsock_param<<16)+
		(sizeof(void *)*4)+(sizeof(int)*3));

	_call_winsock(0x100, &params, sizeof(params));

	if(_kipdj_error && _kipdj_error!=0xffff)
		return(-1);

	/* si no hay socket conectado, error */
	if(params.conn_socket==0)
		return(-1);

	*addrlen=params.len;

	/* recoge el sockaddr */
	memcpy(addr, &params.addr, *addrlen);

	return((int) params.conn_socket);
}


int bind(int s, struct sockaddr * addr, int addrlen)
/* asigna una direccin o puerto al socket local */
{
	struct
	{
		void * address;
		void * socket;
		int len;
		void * routine;
		int context;
		struct sockaddr addr;
	} params;

	params.socket=(void *)s;
	params.len=addrlen;
	params.routine=NULL;
	params.context=0;

	params.address=(void *)((_winsock_param<<16)+
		(sizeof(void *)*3)+(sizeof(int)*2));

	memcpy(&params.addr, addr, addrlen);

	_call_winsock(0x101, &params, sizeof(params));

	memcpy(addr, &params.addr, addrlen);

	if(_kipdj_error)
		return(-1);

	return(0);
}


int getpeername(int s, struct sockaddr * addr, int * len)
/* obtiene informacin sobre el lado contrario del socket */
{
	struct
	{
		void * address;
		void * socket;
		int addrlen;
		struct sockaddr addr;
	} params;

	params.socket=(void *) s;
	params.addrlen=*len;

	params.address=(void *)((_winsock_param<<16)+
		(sizeof(void *)*2)+sizeof(int));

	_call_winsock(0x104, &params, sizeof(params));

	if(_kipdj_error && _kipdj_error!=0xffff)
		return(-1);

	*len=params.addrlen;
	memcpy(addr, &params.addr, *len);

	return(0);
}


int getsockname(int s, struct sockaddr * addr, int * len)
/* obtiene informacin sobre este lado del socket */
{
	struct
	{
		void * address;
		void * socket;
		int addrlen;
		struct sockaddr addr;
	} params;

	params.socket=(void *) s;
	params.addrlen=*len;

	params.address=(void *)((_winsock_param<<16)+
		(sizeof(void *)*2)+sizeof(int));

	_call_winsock(0x105, &params, sizeof(params));

	if(_kipdj_error && _kipdj_error!=0xffff)
		return(-1);

	*len=params.addrlen;
	memcpy(addr, &params.addr, *len);

	return(0);
}


int ioctlsocket(int s, int command, unsigned long * param)
/* ioctl */
{
	struct
	{
		void * socket;
		int command;
		unsigned long param;
	} params;

	if(command==FIONBIO)
	{
		/* si se piden sockets bloqueantes,
		   simplemente se define esta variable
		   tiene algn efecto lateral? */

		_kipdj_blocking=!*param;
		return(0);
	}

	params.socket=(void *)s;
	params.command=command;
	params.param=*param;

	_call_winsock(0x107, &params, sizeof(params));

	if(_kipdj_error)
		return(-1);

	*param=params.param;

	return(0);
}


int getsockopt(int s, int level, int optname,
	void * optval, int * optlen)
/* obtiene el valor de una opcin de socket */
{
	struct
	{
		void * valaddr;
		void * socket;
		int level;
		int optname;
		int optlen;
		int optval;
	} params;
	int * i;

	params.socket=(void *)s;
	params.level=level;
	params.optname=optname;
	params.optlen=*optlen;
	params.optval=0;

	/* hace apuntar valaddr a optval */
	params.valaddr=(void *) ((_winsock_param<<16)+
		(sizeof(void *)*2)+(sizeof(int)*3));

	_call_winsock(0x106, &params, sizeof(params));

	if(_kipdj_error)
		return(-1);

	*optlen=params.optlen;
	i=(int *) optval;
	*i=params.optval;

	return(0);
}


int setsockopt(int s, int level, int optname,
	const void * optval, int optlen)
/* define una opcin de socket */
{
	struct
	{
		void * valaddr;
		void * socket;
		int level;
		int optname;
		int optlen;
		int optval;
	} params;

	params.socket=(void *)s;
	params.level=level;
	params.optname=optname;
	params.optlen=optlen;
	params.optval=*((int *)optval);

	params.valaddr=(void *)((_winsock_param<<16)+
		(sizeof(void *)*2)+(sizeof(int)*3));

	_call_winsock(0x10e, &params, sizeof(params));

	if(_kipdj_error)
		return(-1);

	return(0);
}


int shutdown(int s, int how)
/* corta la comunicacin en un socket */
{
	struct
	{
		void * socket;
		int how;
	} params;

	params.socket=(void *)s;
	params.how=how;

	_call_winsock(0x10f, &params, sizeof(params));

	if(_kipdj_error)
		return(-1);

	return(0);
}


unsigned short htons(unsigned short s)
/* da la vuelta al short */
{
	return((s<<8)|(s>>8));
}


unsigned short ntohs(unsigned short s)
/* da la vuelta al short */
{
	return((s<<8)|(s>>8));
}


unsigned long htonl(unsigned long l)
/* da la vuelta a un long */
{
	return( ((l&0x000000ff)<<24) | ((l&0x0000ff00)<<8) |
		((l&0x00ff0000)>>8)  | ((l&0xff000000)>>24) );
}


unsigned long ntohl(unsigned long l)
/* da la vuelta a un long */
{
	return( ((l&0x000000ff)<<24) | ((l&0x0000ff00)<<8) |
		((l&0x00ff0000)>>8)  | ((l&0xff000000)>>24) );
}


char * inet_ntoa(struct in_addr in)
/* convierte a caracteres la direccin ipaddr */
{
	static char buf[15];
	char * ip;

	ip=(char *) &in;

	sprintf(buf,"%u.%u.%u.%u",(unsigned char) ip[0],
		(unsigned char) ip[1], (unsigned char) ip[2],
		(unsigned char) ip[3]);

	return(buf);
}


unsigned long int inet_addr(char * addr)
/* convierte una direccin en caracteres a un ip */
{
	char tmp[80];
	char * ip0;
	char * ip1;
	char * ip2;
	char * ip3;
	unsigned long l;

	strcpy(tmp,addr);
	ip0=strtok(tmp,".");
	ip1=strtok(NULL,".");
	ip2=strtok(NULL,".");
	ip3=strtok(NULL,".");

	if(ip0==NULL || ip1==NULL || ip2==NULL || ip3==NULL)
		return((unsigned long) -1);

	l=_kipdj_ip(atoi(ip3),atoi(ip2),atoi(ip1),atoi(ip0));

	return(l);
}


char * strerrorsocket(void)
/* devuelve una cadena con el error */
{
	static char _tmp[25];

	sprintf(_tmp,"kipdj error %d",_kipdj_error);

	return(_tmp);
}


struct hostent * gethostbyname(char * name)
/* obtiene el ip de un host (dummy) */
{
	return(NULL);
}


int _kipdj_startup(void)
/* inicializa kipdj */
{
	__dpmi_meminfo _dpmi_winsock_param;
	__dpmi_meminfo _dpmi_winsock_data;

	/* si ya est inicializado, fuera */
	if(_kipdj_up)
		return(1);

	/* obtiene el punto de entrada del VxD loader */
	_vxd_get_entry(_vxd_entry_point, 0x0027);

	/* no hay VxD loader, fuera */
	if(*_vxd_entry_point==0)
	{
		_kipdj_error=1;
		return(0);
	}

	/* carga los dos posibles controladores VxD */
	/* si alguno de los dos se carga correctamente,
	   se activa _kipdj_up */

	_vxd_load_device("WSOCK.VXD");
	_vxd_load_device("WSOCK.386");

	if(!_kipdj_up)
	{
		_kipdj_error=2;
		return(0);
	}

	/* obtiene el punto de entrada de Winsock */
	_vxd_get_entry(_winsock_entry_point, 0x003e);

	if(*_winsock_entry_point==0)
	{
		_kipdj_up=0;
		_kipdj_error=3;
		return(0);
	}

	_dpmi_winsock_param.handle=0;
	_dpmi_winsock_param.size=65536;
	_dpmi_winsock_param.address=0;
	__dpmi_allocate_memory(&_dpmi_winsock_param);

	_dpmi_winsock_data.handle=0;
	_dpmi_winsock_data.size=65536;
	_dpmi_winsock_data.address=0;
	__dpmi_allocate_memory(&_dpmi_winsock_data);

	_winsock_param=__dpmi_allocate_ldt_descriptors(1);
	_winsock_data=__dpmi_allocate_ldt_descriptors(1);

	__dpmi_set_segment_base_address(_winsock_param,
		_dpmi_winsock_param.address);
	__dpmi_set_segment_base_address(_winsock_data,
		_dpmi_winsock_data.address);

	__dpmi_set_segment_limit(_winsock_param, _dpmi_winsock_param.size);
	__dpmi_set_segment_limit(_winsock_data, _dpmi_winsock_data.size);

	return(1);
}


#ifdef DEMO
int main(void)
{
	int s,s2;
	struct sockaddr_in sin;
	char buffer[4096];
	int n;

	printf("0x%08X\n",_ipaddress(111,1,2,1));

	if(!SocketStartup())
	{
		printf("Error en init\n");
		return(0);
	}

	s=socket(AF_INET, SOCK_STREAM, 0);

	printf("s==%d, s2==%d\n",s,s2);

	sin.sin_family=AF_INET;
	sin.sin_port=htons(79); /* 79 */
/*	  sin.sin_addr.s_addr=htonl(0x6f010201);*/
	sin.sin_addr.s_addr=htonl(_ipaddress(111,1,2,1));
/*	  sin.sin_addr.s_addr=htonl(_ipaddress(192,168,0,7));*/

	printf("sin_port==0x%04X, s_addr==0x%08X\n",sin.sin_port, sin.sin_addr.s_addr);

	if(connect(s,(struct sockaddr *) &sin, sizeof(struct sockaddr))==-1)
		printf("Error en connect\n");

	printf("connect... _kipdj_error==%d\n",_kipdj_error);

	printf("sin_port==0x%04X, s_addr==0x%08X\n",sin.sin_port, sin.sin_addr.s_addr);

	_wait_for_event(s, FD_CONNECT);

	printf("FD_CONNECT... _kipdj_error==%d\n",_kipdj_error);

	n=sizeof(struct sockaddr);
	if(getsockname(s,(struct sockaddr *) &sin, &n)==-1)
		printf("Error en getsockname\n");

	printf("getsockname... _kipdj_error==%d\n",_kipdj_error);

	printf("sin_port==0x%04X, s_addr==0x%08X\n",sin.sin_port, sin.sin_addr.s_addr);

	strcpy(buffer,"johannes\r\n");
	send(s,buffer,strlen(buffer),0);

	printf("send.... _kipdj_error==%d\n",_kipdj_error);

	_wait_for_event(s, FD_READ);
	printf("FD_READ... _kipdj_error==%d\n",_kipdj_error);

	n=recv(s,buffer,4096,0);
	printf("recv==%d\n",n);

	printf("buffer: [%s]\n",buffer);

	printf("recv... _kipdj_error==%d\n",_kipdj_error);
	closesocket(s);


	s2=socket(AF_INET, SOCK_STREAM, 0);

	memset(&sin, '\0', sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_port=htons(6969);

	if(bind(s2,(struct sockaddr *) &sin, sizeof(sin))==-1)
	{
		printf("bind... _kipdj_error==%d\n",_kipdj_error);
		printf("Error en bind\n");
		return(1);
	}

	listen(s2, 5);

	printf("listen... _kipdj_error==%d\n",_kipdj_error);

	n=sizeof(struct sockaddr);
	printf("n==%d\n",n);

	s=accept(s2,(struct sockaddr *) &sin, &n);

	printf("accept... _kipdj_error==%d\n",_kipdj_error);

	if(s==-1)
		printf("Error en accept");
	else
	{
		send(s,"mierda\r\n",8,0);
		send(s,"cagao\r\n",7,0);
		closesocket(s);
	}

	closesocket(s2);

	return(0);
}
#endif
