/*

	kipdj.h

	(C) Angel Ortega 1997/1998

*/


/* externas */

extern int _kipdj_up;
extern int _kipdj_error;
extern int _kipdj_blocking;


/* familias de socket */

#define AF_UNSPEC	0
#define AF_UNIX 	1
#define AF_INET 	2
#define AF_IMPLINK	3
#define AF_PUP		4
#define AF_CHAOS	5
#define AF_NS		6
#define AF_ISO		7
#define AF_OSI		AF_ISO
#define AF_ECMA 	8
#define AF_DATAKIT	9
#define AF_CCITT	10
#define AF_SNA		11
#define AF_DECnet	12
#define AF_DLI		13
#define AF_LAT		14
#define AF_HYLINK	15
#define AF_APPLETALK	16
#define AF_NETBIOS	17

#define AF_MAX		18

#define PF_UNSPEC	AF_UNSPEC
#define PF_UNIX 	AF_UNIX
#define PF_INET 	AF_INET
#define PF_IMPLINK	AF_IMPLINK
#define PF_PUP		AF_PUP
#define PF_CHAOS	AF_CHAOS
#define PF_NS		AF_NS
#define PF_ISO		AF_ISO
#define PF_OSI		AF_OSI
#define PF_ECMA 	AF_ECMA
#define PF_DATAKIT	AF_DATAKIT
#define PF_CCITT	AF_CCITT
#define PF_SNA		AF_SNA
#define PF_DECnet	AF_DECnet
#define PF_DLI		AF_DLI
#define PF_LAT		AF_LAT
#define PF_HYLINK	AF_HYLINK
#define PF_APPLETALK	AF_APPLETALK

#define PF_MAX		AF_MAX


/* tipos de socket */

#define SOCK_STREAM	1
#define SOCK_DGRAM	2
#define SOCK_RAW	3
#define SOCK_RDM	4
#define SOCK_SEQPACKET	5


/* protocolos */

#define IPPROTO_IP		0
#define IPPROTO_ICMP		1
#define IPPROTO_GGP		2
#define IPPROTO_TCP		6
#define IPPROTO_PUP		12
#define IPPROTO_UDP		17
#define IPPROTO_IDP		22
#define IPPROTO_ND		77

#define IPPROTO_RAW		255
#define IPPROTO_MAX		256


/** estructuras de control de sockets **/

#define PACKED __attribute__((packed))

struct sockaddr
{
	unsigned short sa_family PACKED;
	char sa_data[14] PACKED;
} PACKED;


struct in_addr
{
	unsigned long s_addr;
} PACKED;


struct sockaddr_in
{
	short sin_family PACKED;
	short sin_port PACKED;
	struct in_addr sin_addr PACKED;
	char sin_zero[8] PACKED;
} PACKED;


/* comandos ioctlsocket */

#define FIONBIO  0x8004667E
#define FIONREAD 0x4004667F
#define FIOASYNC 0x8004667D


/* opciones [s|g]etsockopt */

#define SOL_SOCKET	0xffff

#define SO_DEBUG	0x0001
#define SO_ACCEPTCONN	0x0002
#define SO_REUSEADDR	0x0004
#define SO_KEEPALIVE	0x0008
#define SO_DONTROUTE	0x0010
#define SO_BROADCAST	0x0020
#define SO_USELOOPBACK	0x0040
#define SO_LINGER	0x0080
#define SO_OOBINLINE	0x0100
#define SO_DONTLINGER	(u_int)(~SO_LINGER)
#define SO_SNDBUF	0x1001
#define SO_RCVBUF	0x1002
#define SO_SNDLOWAT	0x1003
#define SO_RCVLOWAT	0x1004
#define SO_SNDTIMEO	0x1005
#define SO_RCVTIMEO	0x1006
#define SO_ERROR	0x1007
#define SO_TYPE 	0x1008
#define TCP_NODELAY	0x0001


/* flags de send/recv */

#define MSG_OOB 	0x1
#define MSG_PEEK	0x2
#define MSG_DONTROUTE	0x4

#define INADDR_ANY	0x0


/* estructura de datos de host */

struct hostent
{
	char * h_name;
	char ** h_aliases;
	int h_addrtype;
	int h_length;
	char ** h_addr_list;
};
#define h_addr h_addr_list[0]


/* macros */

#define _kipdj_ip(a,b,c,d) (((a)<<24)|((b)<<16)|((c)<<8)|(d))


/* prototipos */

int socket     (int domain, int type, int protocol);
int closesocket(int s);
int connect    (int s, struct sockaddr * addr, int len);
int sendto     (int s, void * data, int len, unsigned int flags,
		struct sockaddr * to, int tolen);
int send       (int s, void * data, int len, unsigned int flags);
int recvfrom   (int s, void * data, int len, unsigned int flags,
		struct sockaddr * from, int * fromlen);
int recv       (int s, void * data, int len, unsigned int flags);
int listen     (int s, int backlog);
int accept     (int s, struct sockaddr * addr, int * addrlen);
int bind       (int s, struct sockaddr * addr, int addrlen);
int getpeername(int s, struct sockaddr * addr, int * len);
int getsockname(int s, struct sockaddr * addr, int * len);
int ioctlsocket(int s, int command, unsigned long * param);
int getsockopt (int s, int level, int optname,
		void * optval, int * optlen);
int setsockopt (int s, int level, int optname,
		const void * optval, int optlen);
int shutdown   (int s, int how);

unsigned short htons(unsigned short s);
unsigned short ntohs(unsigned short s);
unsigned long htonl(unsigned long l);
unsigned long ntohl(unsigned long l);

char * inet_ntoa(struct in_addr in);
unsigned long int inet_addr(char * addr);

struct hostent * gethostbyname(char * name);

char * strerrorsocket(void);
int _kipdj_startup(void);

#define __close     closesocket
#define sclose	    closesocket
#define __ioctl     ioctlsocket
#define sioctl	    ioctlsocket

#ifndef EWOULDBLOCK
#define EWOULDBLOCK EAGAIN
#endif
